"""
Mixin for all implementations of :py:class:`RemoteAPI` for the Spotify API.

Also includes the default arguments to be used when requesting authorisation from the Spotify API.
"""
import base64
from collections.abc import Iterable
from copy import deepcopy

from yarl import URL

from musify import PROGRAM_NAME
from musify.api.authorise import APIAuthoriser
from musify.api.cache.backend.base import ResponseCache, ResponseRepository
from musify.api.cache.session import CachedSession
from musify.api.exception import APIError
from musify.libraries.remote.spotify.api.cache import SpotifyRequestSettings, SpotifyPaginatedRequestSettings
from musify.libraries.remote.spotify.api.item import SpotifyAPIItems
from musify.libraries.remote.spotify.api.misc import SpotifyAPIMisc
from musify.libraries.remote.spotify.api.playlist import SpotifyAPIPlaylists
from musify.libraries.remote.spotify.wrangle import SpotifyDataWrangler
from musify.utils import safe_format_map, merge_maps

URL_AUTH = "https://accounts.spotify.com"

# user authenticated access with scopes
SPOTIFY_API_AUTH_ARGS = {
    "auth_args": {
        "url": f"{URL_AUTH}/api/token",
        "data": {
            "grant_type": "authorization_code",
        },
        "headers": {
            "content-type": "application/x-www-form-urlencoded",
            "Authorization": "Basic {client_base64}"
        },
    },
    "user_args": {
        "url": f"{URL_AUTH}/authorize",
        "params": {
            "client_id": "{client_id}",
            "response_type": "code",
            "state": PROGRAM_NAME,
            "scope": "{scopes}",
            "show_dialog": False,
        },
    },
    "refresh_args": {
        "url": f"{URL_AUTH}/api/token",
        "data": {
            "grant_type": "refresh_token",
        },
        "headers": {
            "content-type": "application/x-www-form-urlencoded",
            "Authorization": "Basic {client_base64}"
        },
    },
    "test_args": {"url": "{url}/me"},
    "test_condition": lambda r: SpotifyAPI.url_key in r and "display_name" in r,
    "test_expiry": 600,
    "token_key_path": ["access_token"],
    "header_extra": {"Accept": "application/json", "Content-Type": "application/json"},
}


class SpotifyAPI(SpotifyAPIMisc, SpotifyAPIItems, SpotifyAPIPlaylists):
    """
    Collection of endpoints for the Spotify API.

    :param client_id: The client ID to use when authorising requests.
    :param client_secret: The client secret to use when authorising requests.
    :param scopes: The scopes to request access to.
    :param cache: When given, attempt to use this cache for certain request types before calling the API.
    :param auth_kwargs: Optionally, provide kwargs to use when instantiating the :py:class:`APIAuthoriser`.
    """

    __slots__ = ()

    @property
    def user_id(self) -> str | None:
        """ID of the currently authenticated user"""
        if not self.user_data:
            raise APIError(
                "User data not set. Either set explicitly or enter the "
                f"{self.__class__.__name__} context to set automatically."
            )
        return self.user_data["id"]

    @property
    def user_name(self) -> str | None:
        """Name of the currently authenticated user"""
        if not self.user_data:
            raise APIError(
                "User data not set. Either set explicitly or enter the "
                f"{self.__class__.__name__} context to set automatically."
            )
        return self.user_data["display_name"]

    def __init__(
            self,
            client_id: str | None = None,
            client_secret: str | None = None,
            scopes: Iterable[str] = (),
            cache: ResponseCache | None = None,
            **auth_kwargs
    ):
        wrangler = SpotifyDataWrangler()

        format_map = {
            "client_id": client_id,
            "client_base64": base64.b64encode(f"{client_id}:{client_secret}".encode()).decode(),
            "scopes": " ".join(scopes),
            "url": str(wrangler.url_api)
        }
        auth_kwargs = merge_maps(deepcopy(SPOTIFY_API_AUTH_ARGS), auth_kwargs, extend=False, overwrite=True)
        safe_format_map(auth_kwargs, format_map=format_map)

        auth_kwargs.pop("name", None)
        authoriser = APIAuthoriser(name=wrangler.source, **auth_kwargs)

        super().__init__(authoriser=authoriser, wrangler=wrangler, cache=cache)

    # noinspection PyAsyncCall
    async def _setup_cache(self) -> None:
        if not isinstance(self.handler.session, CachedSession):
            return

        cache = self.handler.session.cache
        cache.repository_getter = self._get_cache_repository

        cache.create_repository(SpotifyRequestSettings(name="tracks"))
        cache.create_repository(SpotifyRequestSettings(name="audio_features"))
        cache.create_repository(SpotifyRequestSettings(name="audio_analysis"))

        cache.create_repository(SpotifyRequestSettings(name="albums"))
        cache.create_repository(SpotifyPaginatedRequestSettings(name="album_tracks"))

        cache.create_repository(SpotifyRequestSettings(name="artists"))
        cache.create_repository(SpotifyPaginatedRequestSettings(name="artist_albums"))

        cache.create_repository(SpotifyRequestSettings(name="shows"))
        cache.create_repository(SpotifyRequestSettings(name="episodes"))
        cache.create_repository(SpotifyPaginatedRequestSettings(name="show_episodes"))

        cache.create_repository(SpotifyRequestSettings(name="audiobooks"))
        cache.create_repository(SpotifyRequestSettings(name="chapters"))
        cache.create_repository(SpotifyPaginatedRequestSettings(name="audiobook_chapters"))

        await cache

    @staticmethod
    def _get_cache_repository(cache: ResponseCache, url: str | URL) -> ResponseRepository | None:
        path = URL(url).path
        path_split = [part.replace("-", "_") for part in path.split("/")[2:]]

        if len(path_split) < 3:
            name = path_split[0]
        else:
            name = "_".join([path_split[0].rstrip("s"), path_split[2].rstrip("s") + "s"])

        return cache.get(name)



"""
The fundamental core enum classes for the entire package.
"""
from enum import IntEnum
from typing import Self

from musify.exception import MusifyEnumError
from musify.utils import unique_list


class MusifyEnum(IntEnum):
    """Generic class for :py:class:`IntEnum` implementations for the entire package."""

    @classmethod
    def map(cls, enum: Self) -> list[Self]:
        """
        Optional mapper to apply to the enum found during :py:meth:`all`, :py:meth:`from_name`,
        and :py:meth:`from_value` calls
        """
        return [enum]

    @classmethod
    def all(cls) -> list[Self]:
        """Get all enums for this enum."""
        return unique_list(e for enum in cls if enum.name != "ALL" for e in cls.map(enum))

    @classmethod
    def from_name(cls, *names: str, fail_on_many: bool = True) -> list[Self]:
        """
        Returns all enums that match the given enum names

        :param fail_on_many: If more than one enum is found, raise an exception.
        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        names_upper = [name.strip().upper() for name in names]
        enums = unique_list(e for enum in cls if enum.name in names_upper for e in cls.map(enum))

        if len(enums) == 0:
            raise MusifyEnumError(names)
        elif len(enums) > 1 and fail_on_many:
            raise MusifyEnumError(value=enums, message="Too many enums found")

        return enums

    @classmethod
    def from_value(cls, *values: int, fail_on_many: bool = True) -> list[Self]:
        """
        Returns all enums that match the given enum values

        :param fail_on_many: If more than one enum is found, raise an exception.
        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        enums = unique_list(e for enum in cls if enum.value in values for e in cls.map(enum))
        if len(enums) == 0:
            raise MusifyEnumError(values)
        elif len(enums) > 1 and fail_on_many:
            raise MusifyEnumError(value=enums, message="Too many enums found")
        return enums



"""
All type hints to use throughout the module.
"""
from collections.abc import MutableMapping
from typing import Any, Mapping

from yarl import URL

from musify.libraries.remote.core import RemoteResponse
from musify.types import UnitMutableSequence, UnitSequence

type APIInputValueSingle[T: RemoteResponse] = str | URL | Mapping[str, Any] | T
type APIInputValueMulti[T: RemoteResponse] = (
        UnitSequence[str] |
        UnitMutableSequence[URL] |
        UnitMutableSequence[MutableMapping[str, Any]] |
        UnitSequence[T]
)



"""
All core type hints to use throughout the entire package.
"""
from collections.abc import Iterable, Sequence, MutableSequence, Collection, Mapping, MutableMapping

type UnitIterable[T] = T | Iterable[T]
type UnitCollection[T] = T | Collection[T]
type UnitSequence[T] = T | Sequence[T]
type UnitMutableSequence[T] = T | MutableSequence[T]
type UnitList[T] = T | list[T]

JSON_VALUE = str | int | float | list | dict | bool | None
JSON = Mapping[str, JSON_VALUE]
MutableJSON = MutableMapping[str, JSON_VALUE]
DictJSON = dict[str, JSON_VALUE]

Number = int | float



import contextlib
from http.client import InvalidURL
from typing import Self, Any

from aiohttp import ClientSession, ClientRequest, payload
from yarl import URL

from musify.api.cache.backend.base import ResponseCache, ResponseRepository
from musify.api.cache.response import CachedResponse

ClientSession.__init_subclass__ = lambda *_, **__: _  # WORKAROUND: disables inheritance warning


class CachedSession(ClientSession):
    """
    A modified session which attempts to get/save responses from/to a stored cache before/after sending it.

    :param cache: The cache to use for managing cached responses.
    """

    __slots__ = ("cache",)

    def __init__(self, cache: ResponseCache, **kwargs):
        super().__init__(**kwargs)

        #: The cache to use when attempting to return a cached response.
        self.cache = cache

    async def __aenter__(self) -> Self:
        self.cache = await self.cache.__aenter__()
        await super().__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await super().__aexit__(exc_type, exc_val, exc_tb)
        await self.cache.__aexit__(exc_type, exc_val, exc_tb)

    @contextlib.asynccontextmanager
    async def request(self, method: str, url: str | URL, json: Any = None, persist: bool = True, **kwargs):
        """
        Perform HTTP request.

        :param method: HTTP request method (such as GET, POST, PUT, etc.)
        :param url: The URL to perform the request on.
        :param json: A JSON serializable Python object to send in the body of the request.
        :param persist: Whether to persist responses returned from sending network requests i.e. non-cached responses.
        :return: Either the :py:class:`CachedResponse` if a response was found in the cache,
            or the :py:class:`ClientResponse` if the request was sent.
        """
        try:
            url = self._build_url(url)
        except ValueError as exc:
            raise InvalidURL(url) from exc

        kwargs["headers"] = kwargs.get("headers", {}) | dict(self.headers)
        if json is not None:
            kwargs["data"] = payload.JsonPayload(json, dumps=self._json_serialize)

        drop_kwargs = ("allow_redirects",)

        req = ClientRequest(
            method=method,
            url=url,
            loop=self._loop,
            response_class=self._response_class,
            session=self,
            trust_env=self.trust_env,
            **{k: v for k, v in kwargs.items() if k not in drop_kwargs},
        )

        repository = self.cache.get_repository_from_requests(req.request_info)
        response = await self._get_cached_response(req, repository=repository)
        if response is None:
            response = await super().request(method=method, url=url, **kwargs)

        yield response

        if persist and repository is not None and not isinstance(response, CachedResponse):
            await repository.save_response(response)

    async def _get_cached_response(
            self, request: ClientRequest, repository: ResponseRepository | None
    ) -> CachedResponse | None:
        if repository is None:
            return

        data = await repository.get_response(request)
        if data is None:
            return

        if not isinstance(data, str | bytes):
            repository = self.cache.get_repository_from_url(request.url)
            data = repository.serialize(data)

        return CachedResponse(request=request, data=data)



from typing import Any

from yarl import URL

from musify.api.cache.backend.base import RequestSettings
from musify.libraries.remote.core.enum import RemoteIDType
from musify.libraries.remote.core.exception import RemoteObjectTypeError
from musify.libraries.remote.spotify.wrangle import SpotifyDataWrangler


class SpotifyRequestSettings(RequestSettings):

    @property
    def fields(self) -> tuple[str, ...]:
        return "id",

    def get_key(self, url: str | URL, *_, **__) -> tuple[str | None, ...]:
        try:
            return SpotifyDataWrangler.convert(str(url), type_in=RemoteIDType.URL, type_out=RemoteIDType.ID),
        except RemoteObjectTypeError:
            pass
        return (None,)

    @staticmethod
    def get_name(response: dict[str, Any]) -> str | None:
        if response.get("type") == "user":
            return response["display_name"]
        return response.get("name")


class SpotifyPaginatedRequestSettings(SpotifyRequestSettings):

    @property
    def fields(self) -> tuple[str, ...]:
        return *super().fields, "offset", "size"

    def get_key(self, url: str | URL, *_, **__) -> tuple[str | int | None, ...]:
        base = super().get_key(url=url)
        return *base, self.get_offset(url), self.get_limit(url)

    @staticmethod
    def get_offset(url: str | URL) -> int:
        """Extracts the offset for a paginated request from the given ``url``."""
        params = URL(url).query
        return int(params.get("offset", 0))

    @staticmethod
    def get_limit(url: str | URL) -> int:
        """Extracts the limit for a paginated request from the given ``url``."""
        params = URL(url).query
        return int(params.get("limit", 50))
